%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id: snmpm.erl,v 1.1.1.1 2004/07/05 09:16:55 tobiasl Exp $
%%
-module(snmpm).

%%----------------------------------------------------------------------
%% This module implements a simple SNMP manager for Erlang.
%%----------------------------------------------------------------------

%% User interface
-export([
	 %% Management API
	 start_link/1, 

	 register_user/3, unregister_user/1, 

	 register_agent/2, register_agent/3, register_agent/4, 
	 unregister_agent/2, unregister_agent/3,
	 
	 update_agent_info/5, agent_info/3,

	 which_agents/0, which_agents/1, 

	 
	 %% 
	 %% Basic SNMP API
	 g/3, g/4, g/5, ag/3, ag/4, ag/5, 
	 gn/3, gn/4, gn/5, agn/3, agn/4, agn/5, 
	 gb/5, gb/6, gb/7, 
	 s/3, s/4, s/5, as/3, as/4, as/5, 
	 cancel_async_request/2, 
	 
	 %% Extended SNMP API
	 discovery/1, discovery/2, discovery/3, discovery/4, 

	 %% Logging
	 log_to_txt/2, log_to_txt/3, log_to_txt/4,
	 change_log_size/1,

	 reconfigure/0,

	 system_start_time/0,
	 sys_up_time/0,

	 verbosity/2 
	]).

%% Application internal export
-export([start_link/3]).


-define(audit_trail_log_name, "snmpm log").
-define(audit_trail_log_file, "snmpm.log").
-define(DEFAULT_AGENT_PORT, 161).


%% This function is called when the snmp application
%% starts. 
start_link(Opts, normal, []) ->
    start_link(Opts).


%% This function is normally not used. Instead the manager is
%% started as a consequence of a call to application:start(snmp)
%% when {snmp, [{services, [{manager, Options}]}]} is present in the
%% node config file.
start_link(Opts) ->
    %% This start the manager top supervisor, which in turn
    %% starts the other processes.
    {ok, _} = snmpm_supervisor:start_link(Opts),
    ok.


%% Change the verbosity of a process in the manager
verbosity(config, V) ->
    snmpm_config:verbosity(V);
verbosity(server, V) ->
    snmpm_server:verbosity(V);
verbosity(net_if, V) ->
    snmpm_server:verbosity(net_if, V);
verbosity(note_store, V) ->
    snmpm_server:verbosity(note_store, V);
verbosity(all, V) ->
    snmpm_config:verbosity(V),
    snmpm_server:verbosity(V),
    snmpm_server:verbosity(net_if, V),
    snmpm_server:verbosity(note_store, V).



%% Register the 'user'. 
%% The manager entity responsible for a specific agent. 
%% Module is the callback module (snmpm_user behaviour) which 
%% will be called whenever something happens (detected 
%% agent, incomming reply or incomming trap/notification).
%% Note that this could have already been done as a 
%% consequence of the node config.
register_user(Id, Module, Data) ->
    snmpm_config:register_user(Id, Module, Data).

unregister_user(Id) ->
    snmpm_config:register_user(Id).


%% Explicitly instruct the manager to handle this agent.
%% Called to instruct the manager that this agent 
%% shall be handled. These functions is used when
%% the user know's in advance which agents the
%% manager shall handle.
%% Note that there is an alternate way to do the same thing:
%% Add the agent to the manager config files.
%% 
%% UserId -> Id of the user responsible for this agent
%% Addr   -> Address of the agent: term()
%% Port   -> Port number of the agent: integer()
%% Config -> Agent configuration: [config()]
register_agent(UserId, Addr) ->
    register_agent(UserId, Addr, ?DEFAULT_AGENT_PORT, []).

register_agent(UserId, Addr, Port) when integer(Port) ->
    register_agent(UserId, Addr, Port, []);
register_agent(UserId, Addr, Config) when list(Config) ->
    register_agent(UserId, Addr, ?DEFAULT_AGENT_PORT, Config).

register_agent(UserId, Addr, Port, Config) ->
    snmpm_config:register_agent(UserId, Addr, Port, Config).

unregister_agent(UserId, Addr) ->
    snmpm_config:unregister_agent(UserId, Addr, ?DEFAULT_AGENT_PORT).

unregister_agent(UserId, Addr, Port) ->
    snmpm_config:unregister_agent(UserId, Addr, Port).

agent_info(Addr, Port, Item) ->
    snmpm_config:agent_info(Addr, Port, Item).

update_agent_info(UserId, Addr, Port, Item, Val) ->
    snmpm_config:update_agent_info(UserId, Addr, Port, Item, Val).

which_agents() ->
    snmpm_config:which_agent().

which_agents(UserId) ->
    snmpm_config:which_agent(UserId).

%% Start a discovery process
discovery(UserId) ->
    snmpm_server:discovery(UserId).

discovery(UserId, AddrOrPort) ->
    snmpm_server:discovery(UserId, AddrOrPort).

% discovery(UserId, Port, Expire) when integer(Port), integer(Expire) ->
%     snmpm_server:discovery(UserId, Port, Expire);
discovery(UserId, Addr, Port) ->
    snmpm_server:discovery(UserId, Addr, Port).

discovery(UserId, Addr, Port, Expire) ->
    snmpm_server:discovery(UserId, Addr, Port, Expire).



%% get request
%% g(Addr, Port, Oids) -> {ok, SnmpReply, Remaining} | {error, Reason}
%% Addr      -> string() | atom | {A1, A2, A3, A4}
%% A1, A2, A3, A4 -> integer()
%% Port      -> integer()
%% SnmpReply -> snmp_reply()
%% Remaining -> integer() (Remaining time of the given or default timeout time)
%% Reason    -> term()
%% 
g(UserId, Addr, Oids) ->
    g(UserId, Addr, ?DEFAULT_AGENT_PORT, Oids).

g(UserId, Addr, Port, Oids) when integer(Port), list(Oids) ->
    snmpm_server:sync_get(UserId, Addr, Port, Oids);
g(UserId, Addr, Oids, Timeout) when list(Oids), integer(Timeout) ->
    g(UserId, Addr, ?DEFAULT_AGENT_PORT, Oids, Timeout).

g(UserId, Addr, Port, Oids, Timeout) ->
    snmpm_server:sync_get(UserId, Addr, Port, Oids, Timeout).


%% asynchroneous get request
%% 
%% ag(UserId, Addr, Oids) -> {ok, ReqId} | {error, Reason}
%% ag(UserId, Addr, Port, Oids) -> {ok, ReqId} | {error, Reason}
%% ag(UserId, Addr, Oids, Expire) -> {ok, ReqId} | {error, Reason}
%% ag(UserId, Addr, Port, Oids, Expire) -> {ok, ReqId} | {error, Reason}
%% UserId -> term()
%% Addr -> address()
%% Port -> integer()
%% Oids -> [oid()]
%% Expire -> integer()
%% ReqId -> term()
%% Reason -> term()
%% 
%% Description:
%% 
%% The reply will be delivered to the user
%% through a call to handle_pdu/5
%% 
ag(UserId, Addr, Oids) ->
    ag(UserId, Addr, ?DEFAULT_AGENT_PORT, Oids).

ag(UserId, Addr, Port, Oids) when integer(Port), list(Oids) ->
    asnmpm_server:async_get(UserId, Addr, Port, Oids);
ag(UserId, Addr, Oids, Expire) when list(Oids), integer(Expire) ->
    ag(UserId, Addr, ?DEFAULT_AGENT_PORT, Oids, Expire).

ag(UserId, Addr, Port, Oids, Expire) ->
    snmpm_server:async_get(UserId, Addr, Port, Oids, Expire).


%% get-next request
gn(UserId, Addr, Oids) ->
    gn(UserId, Addr, ?DEFAULT_AGENT_PORT, Oids).

gn(UserId, Addr, Port, Oids) when integer(Port), list(Oids) ->
    snmpm_server:sync_get_next(UserId, Addr, Port, Oids);
gn(UserId, Addr, Oids, Timeout) when list(Oids), integer(Timeout) ->
    gn(UserId, Addr, ?DEFAULT_AGENT_PORT, Oids, Timeout).

gn(UserId, Addr, Port, Oids, Timeout) ->
    snmpm_server:sync_get_next(UserId, Addr, Port, Oids, Timeout).


agn(UserId, Addr, Oids) ->
    agn(UserId, Addr, ?DEFAULT_AGENT_PORT, Oids).

agn(UserId, Addr, Port, Oids) when integer(Port), list(Oids) ->
    snmpm_server:async_get_next(UserId, Addr, Port, Oids);
agn(UserId, Addr, Oids, Expire) when list(Oids), integer(Expire) ->
    agn(UserId, Addr, ?DEFAULT_AGENT_PORT, Oids, Expire).

agn(UserId, Addr, Port, Oids, Expire) ->
    snmpm_server:async_get_next(UserId, Addr, Port, Oids, Expire).


%% set request
s(UserId, Addr, VarsAndVals) ->
    s(UserId, Addr, ?DEFAULT_AGENT_PORT, VarsAndVals).

s(UserId, Addr, Port, VarsAndVals) when integer(Port), list(VarsAndVals) ->
    snmpm_server:sync_set(UserId, Addr, Port, VarsAndVals);
s(UserId, Addr, VarsAndVals, Timeout) 
  when list(VarsAndVals), integer(Timeout) ->
    s(UserId, Addr, ?DEFAULT_AGENT_PORT, VarsAndVals, Timeout).

s(UserId, Addr, Port, VarsAndVals, Timeout) ->
    snmpm_server:sync_set(UserId, Addr, Port, VarsAndVals, Timeout).

as(UserId, Addr, VarsAndVals) ->
    as(UserId, Addr, ?DEFAULT_AGENT_PORT, VarsAndVals).

as(UserId, Addr, Port, VarsAndVals) when integer(Port), list(VarsAndVals) ->
    snmpm_server:async_set(UserId, Addr, Port, VarsAndVals);
as(UserId, Addr, VarsAndVals, Expire) 
  when list(VarsAndVals), integer(Expire) ->
    as(UserId, Addr, ?DEFAULT_AGENT_PORT, VarsAndVals, Expire).

as(UserId, Addr, Port, VarsAndVals, Expire) ->
    snmpm_server:async_set(UserId, Addr, Port, VarsAndVals, Expire).


gb(UserId, Addr, NonRep, MaxRep, Oids) ->
    gb(UserId, Addr, ?DEFAULT_AGENT_PORT, NonRep, MaxRep, Oids).

gb(UserId, Addr, Port, NonRep, MaxRep, Oids) 
  when integer(Port), integer(NonRep), integer(MaxRep), list(Oids) ->
    snmpm_server:sync_get_bulk(UserId, Addr, Port, NonRep, MaxRep, Oids);
gb(UserId, Addr, NonRep, MaxRep, Oids, Timeout) 
  when integer(NonRep), integer(MaxRep), list(Oids), integer(Timeout) ->
    gb(UserId, Addr, ?DEFAULT_AGENT_PORT, NonRep, MaxRep, Oids, Timeout).

gb(UserId, Addr, Port, NonRep, MaxRep, Oids, Timeout) ->
    snmpm_server:sync_get_bulk(UserId, Addr, Port, NonRep, MaxRep, Oids, Timeout).


cancel_async_request(UserId, ReqId) ->
    snmpm_server:cancel_async_request(UserId, ReqId).


%%%-----------------------------------------------------------------
%%% Audit Trail Log functions (for backward compatibillity)
%%%-----------------------------------------------------------------
log_to_txt(LogDir, Mibs) ->
    OutFile = "snmpm_log.txt",       
    LogName = ?audit_trail_log_name, 
    LogFile = ?audit_trail_log_file, 
    snmp:log_to_txt(LogName, LogFile, LogDir, Mibs, OutFile).
log_to_txt(LogDir, Mibs, OutFile) ->
    LogName = ?audit_trail_log_name, 
    LogFile = ?audit_trail_log_file, 
    snmp:log_to_txt(LogName, LogFile, LogDir, Mibs, OutFile).
log_to_txt(LogDir, Mibs, OutFile, LogName) ->
    LogFile = ?audit_trail_log_file, 
    snmp:log_to_txt(LogName, LogFile, LogDir, Mibs, OutFile).


change_log_size(NewSize) ->
    LogName = ?audit_trail_log_name, 
    snmp:change_size(LogName, NewSize).

reconfigure() ->
    snmpm_server:reconfigure().


%%%-----------------------------------------------------------------

system_start_time() ->
    {ok, Time} = snmpm_config:system_start_time(),
    Time.

sys_up_time() ->
    % time in 0.01 seconds.
    StartTime = system_start_time(),
    (snmp_misc:now(cs) - StartTime) rem (2 bsl 31).


